# pylint: disable=logging-fstring-interpolation
""" Un module pour illustrer les principaux alogorithmes de jointure en Python
    On ne gèrera ici le cas nominal d'utilisation des fonctions où
    attr1 et attr2 sont LES INDEX EXISTANT des attributs
"""
# %%
import argparse
import logging
import operator as op
from collections import defaultdict
from pathlib import Path
from pprint import pprint
from random import randrange, seed
from time import perf_counter
from timeit import timeit

logger = logging.getLogger("JOINS")

sample_table_1 = [
    (27, "Jonah"),
    (28, "Alan"),
    (18, "Alan"),
    (10, "Peppa"),
    (28, "Glory"),
    (18, "Popeye"),
    (19, "Popeye"),
]

sample_table_2 = [
    ("Jonah", "Whales"),
    ("Alan", "Ghosts"),
    ("Jonah", "Spiders"),
    ("Alan", "Zombies"),
    ("Glory", "Buffy"),
    ("John", "Bats"),
    ("Popeye", "Spinach"),
]

# Alan est 2x dans t1 et 2x dans t2, il apparait 4 fois dans la jointure
# Glory est 1x dans t1 et 1x dans t2, elle apparait 1 fois dans la jointure
# Jonah est 1x dans t1 et 2x dans t2, il apparait 2 fois dans la jointure
# Popeye est 2x dans t1 et 1x dans t2, il apparait 2 fois dans la jointure
# Peppa est 1x dans t1 et 0x dans t2, elle apparait 0 fois dans la jointure
# John est 0x dans t1 et 1x dans t2, il apparait 0 fois dans la jointure

reference_result = [
    (18, "Alan", "Alan", "Ghosts"),
    (18, "Alan", "Alan", "Zombies"),
    (28, "Alan", "Alan", "Ghosts"),
    (28, "Alan", "Alan", "Zombies"),
    (28, "Glory", "Glory", "Buffy"),
    (27, "Jonah", "Jonah", "Spiders"),
    (27, "Jonah", "Jonah", "Whales"),
    (18, "Popeye", "Popeye", "Spinach"),
    (19, "Popeye", "Popeye", "Spinach"),
]


def join_nested_loop(table1, attr1, table2, attr2):
    """L'algorithme naif de jointure par deux boucles imbriquées :
    pour chaque tuple de tabl1, on va lire toute la table2
    et produire un résultat à chaque fois que tup1[attr1] == tup2[attr2]
    _!_ on pourrait mettre une condition booléenne arbitraire _!_
    https://en.wikipedia.org/wiki/Nested_loop_join

    CONTRAT: pas de préconditions sur les entrées"""

    for tup1 in table1:
        for tup2 in table2:
            if op.eq(tup1[attr1], tup2[attr2]):
                yield tup1 + tup2


def join_hash(table1, attr1, table2, attr2):
    """L'algorithme de jointure en deux étapes :
      1. construction d'un index sur table1 via un dictionnaire :
         à chaque valeur de attr1, on associe **la liste** des index des
         tuples de table1 qui ont cette valeur
      2. scan de table2 :
         pour chaque tuple, on va chercher AVEC L'INDEX la liste
         des tuples de table1 qui on la valeur de attr2 pour attr1
         on ajoute les tuples au résultat (pas de besoin de vérifier
         les collisions car Python s'en charge pour nous avec le dictionnaire)

    https://en.wikipedia.org/wiki/Hash_join#Classic_hash_join

    CONTRAT: pas de préconditions sur les entrées

    Pour expliciter l'étape 1, sur le jeu d'essai des tests, on
    devrait obtenir le dictionnaire suivant qui dit qu'on peut
    trouver un tuple ayant valeur 'Jonah' à l'index 0 de la table1,
    'Alan' aux index 1 et 2, etc. soit :
        'Jonah': [0],
        'Alan': [1, 2],
        'Peppa': [3],
        'Glory': [4],
        'Popeye': [5, 6]

    https://docs.python.org/3.10/library/collections.html#collections.defaultdict
    Comme on utilise un defaultdict, pour toutes les autres
    clefs le dictionnaire renverra []
    """

    # en fait, comme les dictionnaires Python utilisent des tables de hash
    # il suffit d'un dictionnaire qui sert d'index (un peu comme pour le
    # décorateur memoize) pour implémenter cet algorithme
    hash_table = defaultdict(list)

    # shamelessly inspired from https://rosettacode.org/wiki/Hash_join#Python
    # hash phase : construction de l'index avec la hashMap
    for idx, tup in enumerate(table1):
        hash_table[tup[attr1]].append(idx)

    # join phase
    for tup in table2:
        for idx in hash_table[tup[attr2]]:
            yield table1[idx] + tup


def join_merge(table1, attr1, table2, attr2):
    """L'algorithme de jointure sort-merge qui s'appuie sur des tables SUPPOSEES TRIEES
    https://en.wikipedia.org/wiki/Sort-merge_join
    Son principe est assez similaire à l'étape "merge" du merge sort
    https://en.wikipedia.org/wiki/Merge_sort

    On avance en // sur table1 et table2 avec 2 index ind1 et ind2
     - si table1[ind1][attr1] est avant table2[ind2][attr2]
       on incrémente ind1
     - si table2[ind2][attr2] est avant table1[ind1][attr1]
       on incrémente ind2
     - si les tuples sur lesquels on se trouve respectent la condition
             table1[ind1][attr1] == table2[ind2][attr2]
        alors avec une boucle locale, on va chercher tous les tuples
        de table2 satisfont la condition et ajouter au résultat.
        ensuite on incrémente ind1

    CONTRAT : le trie des entrées est à la charge des utilisateurs,
              le comportement n'est pas garanti sinon"""

    len1, len2 = len(table1), len(table2)
    ind1, ind2 = 0, 0

    # tant que aucun des tableau n'a été parcouru jusqu'au bout
    while ind1 < len1 and ind2 < len2:
        # is les attributs de jointures sont ==
        if table1[ind1][attr1] == table2[ind2][attr2]:
            delta = 0
            # alors on va parcourir tous les tuples de table2 qui ont cette valeur
            while (ind2 + delta < len2) and (
                table1[ind1][attr1] == table2[ind2 + delta][attr2]
            ):
                # et les ajouter au résultat
                yield table1[ind1] + table2[ind2 + delta]
                delta += 1
            # on passe au tuple suivant de table1
            ind1 += 1
        # si l'index de table1 est en retard, on l'avance
        elif table1[ind1][attr1] < table2[ind2][attr2]:
            ind1 += 1
        # sinon, c'est l'index de table2 qui est en retard, on l'avance
        else:
            ind2 += 1


# %%
def benchmark(
    size_1,
    nb_val_1,
    size_2,
    nb_val_2,
    nb_repeat=20,
    bench_loop=True,
    bench_hash=True,
    bench_merge=True,
):  # pylint: disable=too-many-arguments
    """Compare les différentes implémentations"""

    logger.info(
        f"building {size_1} tuples with {nb_val_1} distinct values for table #1"
    )
    sample_1 = [(i, randrange(nb_val_1)) for i in range(size_1)]
    logger.info(
        f"building {size_2} tuples with {nb_val_2} distinct values for table #2"
    )
    sample_2 = [(randrange(nb_val_2), "A" + str(j)) for j in range(size_2)]

    logger.info(f"running each algorithm {nb_repeat} times")
    logger.debug(f"started at {perf_counter()}")
    logger.info(
        f"result contains {sum(1 for _ in join_hash(sample_1, 1, sample_2, 0))} tuples"
    )

    if bench_loop:
        time_loop = timeit(
            lambda: list(join_nested_loop(sample_1, 1, sample_2, 0)), number=nb_repeat
        )
        print(
            f"nested loop : {1000*time_loop/nb_repeat:>6.2f} ms / loop (total {time_loop:.2f} s)"
        )

    if bench_hash:
        time_hash = timeit(
            lambda: list(join_hash(sample_1, 1, sample_2, 0)), number=nb_repeat
        )
        print(
            f"hash        : {1000*time_hash/nb_repeat:>6.2f} ms / loop (total {time_hash:.2f} s)"
        )

    if bench_merge:
        # on utilise la fonction générique sorted() qui produit une nouvelle
        # liste et laisse le paramètre inchangé
        time_sort = timeit(
            lambda: sorted(sample_1, key=lambda x: x[1]), number=nb_repeat
        )
        time_sort += timeit(
            lambda: sorted(sample_2, key=lambda x: x[0]), number=nb_repeat
        )

        # on va trier une bonne fois pour toutes pour que l'algo fonctionne
        sample_1.sort(key=lambda x: x[1])
        sample_2.sort(key=lambda x: x[0])
        time_merge = timeit(
            lambda: list(join_merge(sample_1, 1, sample_2, 0)), number=nb_repeat
        )
        print(
            f"sort        : {1000*time_sort/nb_repeat:>6.2f} ms / loop (total {time_sort:.2f} s)"
        )
        print(
            f"merge       : {1000*time_merge/nb_repeat:>6.2f} ms / loop (total {time_merge:.2f} s)"
        )

    logger.debug(f"ended at   {perf_counter()}")


def get_parser():
    """Configuration de argparse pour les options de ligne de commandes"""
    res = argparse.ArgumentParser(
        prog=f"python {Path(__file__).name}",
        description="Comparaison des performances des algorithmes de jointure (nested loop, sort-merge et hash join).",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )

    res.add_argument(
        "--verbose",
        "-v",
        action="count",
        default=0,
        help="niveau de verbosité, -v pour INFO, -vv pour DEBUG. WARNING par défaut",
    )

    res.add_argument(
        "--demo",
        "-d",
        action="store_true",
        help="affiche une démonstration au lieu du bench",
    )

    res.add_argument(
        "--size-1",
        "-s1",
        action="store",
        default=1000,
        type=int,
        help="nombre de tuples de la première table",
    )

    res.add_argument(
        "--size-2",
        "-s2",
        action="store",
        default=1000,
        type=int,
        help="nombre de tuples de la deuxième table",
    )

    res.add_argument(
        "--vals-1",
        "-v1",
        action="store",
        default=100,
        type=int,
        help="nombre de valeurs sur l'attribut de jointure de la première table",
    )

    res.add_argument(
        "--vals-2",
        "-v2",
        action="store",
        default=100,
        type=int,
        help="nombre de valeurs sur l'attribut de jointure de la deuxième table",
    )

    res.add_argument(
        "--repeat",
        "-r",
        action="store",
        default=20,
        type=int,
        help="nombre de répétitions",
    )

    res.add_argument(
        "--seed",
        "-s",
        action="store",
        default=None,
        type=int,
        help="graine du générateur aléatoire",
    )

    return res


def main():
    """Main entry point"""
    parser = get_parser()
    args = parser.parse_args()

    if args.verbose == 1:
        debug_level = logging.INFO
    elif args.verbose >= 2:
        debug_level = logging.DEBUG
    else:
        debug_level = logging.WARNING

    logging.basicConfig(level=debug_level)
    logger.debug(args)

    if args.seed is not None:
        seed(args.seed)

    if args.demo:
        print("Table 1 (age, name)".center(40, "-"))
        pprint(sample_table_1)
        print("Table 2 (name, monster)".center(40, "-"))
        pprint(sample_table_2)
        print("Résultat attendu (à l'ordre près)".center(40, "-"))
        pprint(reference_result)
        # nécessaire pour sort-merge
        sample_table_1.sort(key=lambda x: x[1])
        sample_table_2.sort(key=lambda x: x[0])
        for algo in [join_nested_loop, join_hash, join_merge]:
            results = list(algo(sample_table_1, 1, sample_table_2, 0))
            # nécessaire pour la comparaison, on trie selon (name, age, monster)
            results.sort(key=lambda x: (x[1], x[0], x[3]))
            assert results == reference_result
            print(f"Résultat de {algo.__name__} correct")

    else:
        benchmark(
            size_1=args.size_1,
            nb_val_1=args.vals_1,
            size_2=args.size_2,
            nb_val_2=args.vals_2,
            nb_repeat=args.repeat,
            bench_loop=True,
            bench_hash=True,
            bench_merge=True,
        )


if __name__ == "__main__":
    main()
